import numpy as np
import casadi as ca
from casadi.tools import *
import matplotlib.pyplot as plt

np.random.seed(927) # Seed random number generator.

# Multiple shooting and integrator parameters
Delta = 0.25 # MHE sampling time
nRK = 10     # number of RK4 steps
nSim = 80    # simulation length

# Plant parameter values
RT = 32.84 # mol atm / L
kp1 = 0.5
km1 = 0.05
kp2 = 0.2
km2 = 0.01

# Standard deviations
sigma_v = 0.25  # Standard deviation of the measurements
sigma_w = 0.001 # Standard deviation for the process noise
sigma_p = 0.5   # Standard deviation for prior

# Make covariance matrices.
P = np.diag((sigma_p*np.ones((3,)))**2) # Covariance for prior.
Q = np.diag((sigma_w*np.ones((1,)))**2)
R = np.diag((sigma_v*np.ones((1,)))**2)

Pinv = np.linalg.inv(P)
Qinv = np.linalg.inv(Q)
Rinv = np.linalg.inv(R)

# Initial state estimate
x0_bar = np.array([1.0,0.0,4.0])

# Real initial state
x0 = np.array([0.5,0.05,0.0])

# Declare the states
x = struct_symSX( ['cA','cB','cC'] ) # Concentration of A, B, C
[cA,cB,cC] = x[...]

# Declare the state disturbance
w = struct_symSX( ['wA','wB','wC'] ) # Disturbance on the system dynamics for A, B, C
[wA,wB,wC] = w[...]

# Declare the measurement
m = struct_symSX( ['y_m'] )
[y] = m[...]

# Declare the measurement disturbance
vy = struct_symSX( ['v'] ) # Disturbance on the measurement y
[v] = vy[...]

# ODE right hand side derivatives
int1 = kp1*cA - km1*cB*cC
int2 = kp2*cB**2 - km2*cC

cA_dot = -int1
cB_dot =  int1 - 2*int2
cC_dot =  int1 + int2
x_dot = ca.vertcat((cA_dot, cB_dot, cC_dot))

# Continuous-time dynamics
f = ca.SXFunction([x], [x_dot])
f.init()

# Discretize the dynamics (RK4 integrator)
dt = Delta/nRK # integrator time step

[k1] = f([x])
[k2] = f([x+0.5*dt*k1])
[k3] = f([x+0.5*dt*k2])
[k4] = f([x+dt*k3])

rk4_step = SXFunction([x],[x + (1./6)*dt*(k1 + 2*k2 + 2*k3 + k4)])
rk4_step.init()

x_next = x
for i in range(nRK):
  [x_next] = rk4_step([x_next])

# Discrete time system
f_dt = SXFunction([x,w],[x_next + w ])
f_dt.init()

# measurement function
measurement = SXFunction([x,v],[ RT*(cA+cB+cC) + v ])
measurement.init()


# First simulate everything
wSim = sigma_w*np.random.randn(nSim,1)
vSim = sigma_v*np.random.randn(nSim,1)

xSim = [x0]
ySim = []
tSim = [0]
for time in range(nSim):
  [xk1] = f_dt([xSim[-1],wSim[time]])
  [yk] = measurement([xSim[-1],vSim[time]])
  xSim.append( np.array(xk1) )
  ySim.append( yk )
  tSim.append( tSim[-1]+Delta )

xSim = horzcat(xSim)
xSim = np.array(xSim.T)

xHat = []
# MHE loop
for time in range(nSim):

  N = time+1 # Full information: the horizon keeps increasing!

  # Multiple Shooting Discretization
  V = struct_symMX([
                    (
                     entry("x",struct=x,repeat=N+1),
                     entry("w",struct=w,repeat=N),
                     entry("v",struct=v,repeat=N),
                     ),
                    ])

  # Constraint vectors (slide 63)
  dynamic_constraints = []
  measurement_constraints = []
  # Penalize the deviation of the initial state from its estimate
  Cost = mul( [ (V['x',0]-x0_bar).T, Pinv, (V['x',0]-x0_bar) ] )
  for k in range(N):
    [xk1] = f_dt([V['x',k],V['w',k]])
    [yk] = measurement([V['x',k],V['v',k]])
    
    # Append the constraints to the constraint vector (slide 63)
    dynamic_constraints.append( xk1 - V['x',k+1] )
    measurement_constraints.append( ySim[k] - yk )

    # Penalize state and measurement noise
    Cost += mul( [ V['w',k].T, Qinv, V['w',k] ] )
    Cost += mul( [ V['v',k].T, Rinv, V['v',k] ] )

  g = struct_MX( [
                  entry("dynamic_constraints",expr=dynamic_constraints),
                  entry("measurement_constraints",expr=measurement_constraints),
                  ] )
  lbg = ubg = g(0)
 
  nlp = MXFunction(nlpIn(x=V),nlpOut(f=Cost,g=g))
  nlp.init()

  # allocate solver
  solver = NlpSolver("ipopt",nlp)

  solver.init()


  V_guess = V()
  # if time == 0:
  #   V_guess['x',0] = x0_bar

  lbx = V(-inf)
  lbx['x'] = 0
  ubx = V(inf)

  solver.setInput(V_guess,  "x0")
  solver.setInput(lbx,  "lbx")
  solver.setInput(ubx,  "ubx")
  solver.setInput(lbg,  "lbg")
  solver.setInput(ubg,  "ubg")

  solver.evaluate()

  Vopt = V(solver.output('x'))
  
  xHat.append( np.squeeze(Vopt['x',-1]) )

xHat = np.array(xHat)

plt.ion()
plt.figure(1)
plt.clf()
# plt.plot(tSim,xSimNoNoise[:,0],'b')
plt.plot(tSim,xSim[:,0],'b')
plt.plot(tSim[:-1],xHat[:,0],'.b')
# plt.plot(tSim,xSimNoNoise[:,1],'r')
plt.plot(tSim,xSim[:,1],'r')
plt.plot(tSim[:-1],xHat[:,1],'.r')
# plt.plot(tSim,xSimNoNoise[:,2],'k')
plt.plot(tSim,xSim[:,2],'k')
plt.plot(tSim[:-1],xHat[:,2],'.k')
plt.grid('on')

plt.figure(2)
plt.clf()
plt.plot(tSim[:-1],ySim,'b')
# plt.plot(tSim[:-1],np.array(ySim)-np.squeeze(vSim),'k')
plt.plot(tSim[:-1],RT*(xHat[:,0]+xHat[:,1]+xHat[:,2]),'.b')
















